package DX::ShellSession;

use Tcl;
use Scalar::Util qw(weaken);
use DX::Expander;
use DX::Proposition;
use DX::Utils qw(deparse);
use DX::Class;

has shell_state => (is => 'rwp', required => 1, isa => ShellState);

has expander => (
  is => 'lazy', builder => sub { DX::Expander->new(tcl => $_[0]->tcl) },
  handles => [ qw(expand_args) ],
);

has tcl => (is => 'lazy', builder => sub {
  my ($self) = @_;
  weaken $self;
  my $tcl = Tcl->new;
  $tcl->CreateCommand('?' => sub {
    $self->apply_to_state([ 'new_query_state' ], [ mode => 'query' ]);
  });
  $tcl->CreateCommand('?+' => sub {
    $self->apply_to_state([ mode => 'query' ]);
  });
  my $qvars;
  $tcl->CreateCommand('.' => sub {
    $self->apply_to_state([ mode => 'shell' ]);
    $qvars->();
  });
  $tcl->CreateCommand(qlist => sub {
    push our @Result, map [ output => $_ ], @{
      $self->shell_state->current_query_state->proposition_sequence->members
    };
    return;
  });
  $tcl->CreateCommand(qvars => $qvars = sub {
    my $locals = $self->shell_state->current_query_state->search_state
                      ->current_hypothesis->scope->locals->[0];
    push our @Result, [ output => $locals ];
    return;
  });
  $tcl->CreateCommand(qdeps => sub {
    my $rps = $self->shell_state->current_query_state->search_state
                   ->current_hypothesis->resolved_propositions;
    push our @Result, [ output => $rps ];
    return;
  });
  foreach my $pred (
    keys %{$self->shell_state->template_query_state->predicates}
  ) {
    $tcl->CreateCommand($pred => sub {
      my (undef, undef, undef, @args) = @_; # ($data, $interp, $pred)
      $self->apply_predicate($pred => @args);
    });
  }
  return $tcl;
});

sub apply_to_state {
  my ($self, @to_apply) = @_;
  my $state = $self->shell_state;
  our @Result;
  foreach my $to_apply (@to_apply) {
    my ($change, @args) = @$to_apply;
    ($state, my @this_result) = $state->${\"with_${change}"}(@args);
    push @Result, @this_result;
  }
  $self->_set_shell_state($state);
  return;
}

sub is_complete_command_string {
  my ($self, $string) = @_;
  return !!$self->tcl->icall(info => complete => $string);
}

sub eval_command_string {
  my ($self, $string) = @_;
  local our @Result;
  try {
    $self->tcl->Eval($string);
  } catch {
    push @Result, [ output => $_ ];
  };
  return map {
    ($_->[0] eq 'output' and ref($_->[1]))
      ? [ output => deparse($_->[1]) ]
      : $_
  } @Result;
}

sub apply_predicate {
  my ($self, $pred, @arg_strings) = @_;
  die "Can't call predicate ${pred} outside a query\n"
    unless $self->shell_state->mode eq 'query';
  my @args = $self->expand_args(@arg_strings);
  my ($intro, $need) = ({}, {});
  foreach my $arg (@args) {
    next if ref($arg);
    # ?Foo is intro, Foo is need
    ($arg =~ s/^\?// ? $intro : $need)->{$arg} = 1;
  }
  my $prop = DX::Proposition->new(
    predicate => $pred,
    args => \@args,
    introduces_names => $intro,
    requires_names => $need,
  );
  my $old_qstate = $self->shell_state->current_query_state;
  my $qstate = $old_qstate->with_additional_proposition($prop);
  my $old_action_count = @{
    $old_qstate->search_state->current_hypothesis->actions
  };
  my @actions = @{$qstate->search_state->current_hypothesis->actions};
  push our @Result,
    map [ output => $_ ], @actions[$old_action_count..$#actions];
  $self->_set_shell_state(
    $self->shell_state->but(
      current_query_state => $qstate
    )
  );
  return;
}

1;
